/*=========================================================================                                                                               
  Program:   bbtk
  Module:    $RCSfile: bbtkInterpreter.h,v $
  Language:  C++
  Date:      $Date: 2008/10/17 08:18:13 $
  Version:   $Revision: 1.35 $
=========================================================================*/

/* ---------------------------------------------------------------------

* Copyright (c) CREATIS-LRMN (Centre de Recherche en Imagerie Medicale)
* Authors : Eduardo Davila, Laurent Guigues, Jean-Pierre Roux
*
*  This software is governed by the CeCILL-B license under French law and 
*  abiding by the rules of distribution of free software. You can  use, 
*  modify and/ or redistribute the software under the terms of the CeCILL-B 
*  license as circulated by CEA, CNRS and INRIA at the following URL 
*  http://www.cecill.info/licences/Licence_CeCILL-B_V1-en.html 
*  or in the file LICENSE.txt.
*
*  As a counterpart to the access to the source code and  rights to copy,
*  modify and redistribute granted by the license, users are provided only
*  with a limited warranty  and the software's author,  the holder of the
*  economic rights,  and the successive licensors  have only  limited
*  liability. 
*
*  The fact that you are presently reading this means that you have had
*  knowledge of the CeCILL-B license and that you accept its terms.
* ------------------------------------------------------------------------ */                                                                         

/**
 *  \file 
 *  \brief class Interpreter : The bbtk language interpreter
 */
/**
 *  \class bbtk::Interpreter 
 *  \brief The bbtk language interpreter
 */

#ifndef __bbtkInterpreter_h__
#define __bbtkInterpreter_h__

#include "bbtkVirtualExec.h"
#include "bbtkExecuter.h"

#include <fstream>
#include <deque>

namespace bbtk
{

#ifdef _USE_WXWIDGETS_
  class WxConsole;
#endif
  
  //=======================================================================
  class BBTK_EXPORT InterpreterUser
  {
  public: 
    InterpreterUser() {}
    virtual ~InterpreterUser() {}

    virtual bool InterpreterUserHasOwnHtmlPageViewer() { return false; }
    virtual void InterpreterUserViewHtmlPage(const std::string&) {}


  };
  //=======================================================================



  //=======================================================================
  class BBTK_EXPORT InterpreterError : public Exception
  {
  public:
    InterpreterError( const std::string& message,
		      bool in_script_file,
		      const std::string& script_file,
		      int script_line 
		      );
    InterpreterError( const Exception& excep,
		      bool in_script_file,
		      const std::string& script_file,
		      int script_line 
		      );
    ~InterpreterError() throw() {}

    bool IsInScriptFile() const { return mInScriptFile; }
    const std::string& GetScriptFile() const { return mScriptFile; }
    int GetScriptLine() const { return mScriptLine; }
  private:
    bool mInScriptFile;
    std::string mScriptFile;
    int mScriptLine;
  };
  //=======================================================================


  //=======================================================================
  class BBTK_EXPORT Interpreter : public Object
  {
    BBTK_OBJECT_INTERFACE_NO_CONDES(Interpreter);
    typedef Object Superclass;
  public:
    static Pointer New(const std::string& cpp_file = "");
    static Pointer New(VirtualExec::Pointer);

    typedef enum 
      {
	Interpreter_OK,
	Interpreter_ERROR,
	Interpreter_QUIT
      }
      ExitStatus;

    /// Runs the interpretation of a file
    ExitStatus InterpretFile( const std::string& filename, bool source = false);

    /// Runs the interpretation of a buffer and deletes it !
    ExitStatus InterpretBuffer( std::stringstream* buffer );

    /// Runs the interpretation of a command
    ExitStatus InterpretLine( const std::string& line );


    /// Launches a command line interpreter (with a prompt)
    void CommandLineInterpreter();



    /// Sets the inputs of the workspace :
    /// the map is passed as is to the Executer
    void SetInputs(const std::map<std::string,std::string>& m)
    { mVirtualExecuter->SetInputs(m); }

    /// Puts the executer in "no exec" mode,
    /// which creates but does not execute pipelines.
    void SetNoExecMode(bool m) { mVirtualExecuter->SetNoExecMode(m); }

    /// Puts the executer in "no error" mode, 
    /// Errors do not stop execution (but warnings are produced)
    void SetNoErrorMode(bool m) { mVirtualExecuter->SetNoErrorMode(m);}
    ///
    //typedef Executer::DialogModeType DialogModeType;
    typedef VirtualExec::DialogModeType DialogModeType;

    void SetDialogMode(DialogModeType t) { mVirtualExecuter->SetDialogMode(t);}

    /// Sets the bool that indicates wether we are in command line context
    void SetCommandLine(bool v = true) { mCommandLine = v; }

    void SetThrow(bool b) { mThrow = b; }

#ifdef _USE_WXWIDGETS_
    /// Sets the user of the interpreter (if any)
    void SetUser(InterpreterUser* c) { mUser = c; }
    /// Gets the InterpreterUser of this 
    InterpreterUser* GetUser() { return mUser; }
    /// Gets the InterpreterUser of this (const)
    const InterpreterUser* GetUser() const { return mUser; }

#endif

    /// Gets the Executer 
    VirtualExec::Pointer GetExecuter() const { return mVirtualExecuter; }


  protected:

    /// The enumeration of command codes == Command name
    typedef enum
    {
      cNew,
      cDelete,
      cConnect,
      cExec,
      cPackage,
      cEndPackage,
      cDefine,
      cEndDefine,
      cInput,
      cOutput,
      cSet,
      cConfig,  // JPR
      cReset,   // EED
      cAuthor, 
      cCategory, // JPR
      cDescription,
      cHelp,
      cMessage,
      cInclude,
      cQuit,
      cLoad,
      cUnload,
      cGraph,
      cPrint,
      cIndex,
      cKind, // LG
      cNewGUI, // LG
      cWorkspace, // LG
      cDebug // LG
    } CommandCodeType;

    /// The structure storing the informations on a command 
    typedef struct 
    {
      std::string keyword;
      int argmin, argmax;
      CommandCodeType code;
      std::string syntax;
      std::string help;
    } CommandInfoType;

    /// The type of dictionnary of commands 
    typedef std::map<std::string,CommandInfoType> CommandDictType;


    /// Interprets a line 
    void InterpretLine( const std::string& line, bool &insideComment );

    /// Reads a line from prompt
    void GetLineFromPrompt( std::string& line );

    /// Splits a line into words
    void SplitLine ( const std::string& line,
                     std::vector<std::string>& words );

    /// Executes the right action depending on the command name
    void InterpretCommand( const std::vector<std::string>& words,
                           CommandInfoType& info );

    /// Switch to the interpretation of a file
    void SwitchToFile( const std::string& filename, bool source = false );

   /// Switch to the interpretation of a stringstream
    void SwitchToStream( std::stringstream* stream );

    /// Closes the currently open file
    void CloseCurrentFile();

    /// Closes all open files
    void CloseAllFiles();

    /// Resets all
    void Reset();

    /// Displays help (entry point of any help)
    void Help(const std::vector<std::string>& words);

    /// Displays help on all the commands
    void HelpCommands();

    /// Displays help on a particular command 
    void HelpCommand( const std::string& command );

    ///
    void Graph(const std::vector<std::string>& words);

    ///
    void Config() const; 
    /// 
    //    void Print(const std::string&);

    void Index(const std::string& filename, 
	       const std::string& type = "Initials");
    ///
    void FindCommandsWithPrefix( char* buf,
				 int n,
				 std::vector<std::string>& commands );

    /// Creates and connects the piece of pipeline which defines a GUI 
    /// for the box box. 
    /// Define it as a complex box type with name instanceName+"Type"
    /// The instance is called instanceName
    /// and connected to the existing pipeline
    void NewGUI(const std::string& box, 
		const std::string& instanceName);

    void Debug(const std::string& arg);

  private:

    /// Constructor
    Interpreter(const std::string& cpp_file = "");

    Interpreter(VirtualExec::Pointer);

    void Init(VirtualExec::Pointer, const std::string& cpp_file);

    /// Opens the file fullPathScriptName 
    /// includeScriptName is the name as given to the include command 
    void LoadScript( std::string fullPathScriptName,
		     std::string includeScriptName);

    /// 
    void CatchBbtkException( const bbtk::Exception& e );
    void CatchStdException( const std::exception& e );
    void CatchUnknownException();


  
    //==================================================================
    // ATTRIBUTES

    /// The command executer used
    bbtk::VirtualExec::Pointer mVirtualExecuter;
    bbtk::Executer::WeakPointer mRealExecuter;

    /// The user of  the interpreter (0 if none)
    bbtk::InterpreterUser* mUser;

    /// Vector of open files / buffers (can be stringstream)
    std::vector<std::istream*> mFile;

    /// Vector of names of open files with full path (as open)
    std::vector<std::string> mFileName;

    /// Vector of names of files which have been open 
    /// before (and may closed then which are no more in mFileName)
    /// with full path (as open)
    std::vector<std::string> mFileNameHistory;

    /// Vector of names of open files as given to the include command
    std::vector<std::string> mIncludeFileName;

    /// Stores the current line number in each open file
    std::vector<int> mLine;

    /// The dictionnary of commands
    CommandDictType mCommandDict;

    /// Are we in a command line context ?
    bool mCommandLine;

    /// The history of commands
    std::deque< char* > mHistory;

   int bufferNb;
    
    bool mThrow;

	protected:
	~Interpreter();

  };
  // EO class Interpreter


  // The "Quit" exception
  struct QuitException : public bbtk::Exception
  {
    QuitException() : bbtk::Exception("Interpreter","","Quit") {}
  };

}
#endif
